/////////////////////////////////////////////////////////////////////
// UVA TeamName RoboCup agent developed for the Simulator
// League of the RoboCup 2001 competition.
//
// This team based on the Mainz Rolling Brains RoboCup team from
// the 2000 competition.
//
// Copyright (C) 2001 Team Members
// of the UVA TeamName RoboCup Team.
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// mastercontrol.cpp Fri Jun 8 2001
// Description:  MasterControl is the "consciousness" of the player
//				 it connects the player to the field and runs the
//				 player through the steps of listening, thinking,
//				 and acting.  See function play() for more information.
//////////////////////////////////////////////////////////////////////

#include "mastercontrol.h"
#include <cstring>
#include "recommendationlist.h"
#include "scanbehavior.h"
#include "interceptballbehavior.h"
#include "shootbehavior.h"
#include "gotogoalbehavior.h"
#include "centerbehavior.h"
#include "dispersebehavior.h"
#include "gohomebehavior.h"
#include "collisionbehavior.h"
#include "kickoffbehavior.h"
#include "stayonfieldbehavior.h"
#include "behindballbehavior.h"
#include "backlinebehavior.h"
#include "clearballbehavior.h"
#include "catchbehavior.h"
#include "goaliepositionbehavior.h"
#include "freekickbehavior.h"
#include "crossbehavior.h"
#include "cornerkickbehavior.h"
#include "puntbehavior.h"
#include "audioworldbehavior.h"
#include "coverbehavior.h"
#include "offensecenterbehavior.h"
#include "gotocornerbehavior.h"
#include "goalieinterceptbehavior.h"
#include "goalkickstallbehavior.h"
#include "cornerpushbehavior.h"

using namespace std;

MasterControl::MasterControl(char* _pcTeamname, const char* _pcHostName, int _iPort, bool _fLogging,
                             int _iReconnectNumber, bool _fIsGoalie)
{
  iLastKickTime = UNKNOWN;
  fClockwise = false;
  
  iLastTime = -1;
  pmLastPlayMode = PLAY_MODE_UNKNOWN;
  
  fLogging = _fLogging;
  // prepare the identity data and register with the server
  IdentityData idAboutMe;
  
  // team_name will be deleted by BaseWorldModel
  // TODO: come up with better way to manage this, perhaps each player should store the teamname
  //			 perhaps it should be a global constant?
  // host information passed in from above, either from command line or as defaults
  idAboutMe.team_name = new char[strlen(_pcTeamname)];
  strcpy(idAboutMe.team_name, _pcTeamname);
  
  if(_fIsGoalie) idAboutMe.type = CLIENT_GOALIE;
  else idAboutMe.type = CLIENT_FIELDPLAYER;
  
  // host information passed in from above, either from command line or as defaults
  bdBody.init(idAboutMe, _pcHostName, _iPort, _iReconnectNumber);
  
  // insert the IdentityData into the WorldModel
  idAboutMe.worldmodel_insert(wmWorld);
  
  // initialize behaviors and perceptions
  if (idAboutMe.number == 11)
    {
      pbeBehaviors.push_back(new OffenseCenterBehavior(wmWorld));
    }
  else if (idAboutMe.number >= 9)
    {
      pbeBehaviors.push_back(new CornerPushBehavior (wmWorld));
    }
  
  if (idAboutMe.number <= 4)
    {
      pbeBehaviors.push_back(new CoverBehavior(wmWorld));
    }

  if(! _fIsGoalie)
    {
      pbeBehaviors.push_back(new ShootBehavior(wmWorld));
      pbeBehaviors.push_back(new GoToGoalBehavior(wmWorld));
      pbeBehaviors.push_back(new GoToCornerBehavior(wmWorld));
      pbeBehaviors.push_back(new FreeKickBehavior(wmWorld));	    
      
      //!pbeBehaviors.push_back(new CornerKickBehavior(wmWorld));
      
      pbeBehaviors.push_back(new CrossBehavior(wmWorld));
      pbeBehaviors.push_back(new DisperseBehavior(wmWorld));
      
      // pbeBehaviors.push_back(new CollisionBehavior(wmWorld));
      pbeBehaviors.push_back(new BackLineBehavior(wmWorld));
      pbeBehaviors.push_back(new BehindBallBehavior(wmWorld));    	
      //      pbeBehaviors.push_back(new CoverBehavior(wmWorld));
      pbeBehaviors.push_back(new CenterBehavior(wmWorld));
      pbeBehaviors.push_back(new InterceptBallBehavior(wmWorld));	
    }
  
  pbeBehaviors.push_back(new ScanBehavior(wmWorld));
  pbeBehaviors.push_back(new AudioWorldBehavior(wmWorld));
  pbeBehaviors.push_back(new KickoffBehavior(wmWorld));
  pbeBehaviors.push_back(new StayOnFieldBehavior(wmWorld));	
  pbeBehaviors.push_back(new ClearBallBehavior(wmWorld));	
  
  if(_fIsGoalie)
    {
      pbeBehaviors.push_back(new GoalKickStallBehavior (wmWorld));
      pbeBehaviors.push_back(new CatchBehavior(wmWorld));
      pbeBehaviors.push_back(new PuntBehavior(wmWorld));	
      pbeBehaviors.push_back(new GoaliePositionBehavior(wmWorld));
      pbeBehaviors.push_back(new GoalieInterceptBehavior(wmWorld));

      // Replaced by GoaliePositionBehavior:
      // pbeBehaviors.push_back(new GoHomeBehavior(wmWorld, Vector2d(-46.5, 0.0) ));		
    }
}

MasterControl::~MasterControl()
{

	unsigned int i;
	for(i = 0; i < pbeBehaviors.size(); i++)
	{
		if (pbeBehaviors[i] != NULL)
			delete pbeBehaviors[i];
	}
}

void MasterControl::play()
{
	if(fLogging)
	{
    	// open log for writing
    	char cFilename[80];
    	snprintf(cFilename, 80, "%s%s%d%s", WORLD_LOG_FILENAME, wmWorld.getTeamName(),
               wmWorld.getClientPlayer().getNumber(), WORLD_LOG_EXTENSION);
     	out.open(cFilename, ios::out);		

     	out << "<?xml version=\"1.0\"?>" << endl;
        out << "<!DOCTYPE playerlog SYSTEM \"playerlog.dtd\">" << endl;

        out << "<playerlog>" << endl;
    }
	
	Recommendation recFinal;
//	int lastTurn = -1;
	
	// Telling the client player move to an initial position
	MoveAction maMove(KICKOFF_POSITIONS[wmWorld.getClientPlayer().getNumber()-1]);
	bdBody.execute(&maMove);
	
	// while the client's body is connected to the server
	while(bdBody.isConnected())
	{		
//		INSPECT(wmWorld.getTime() << "MC: Face b/f listen: " << wmWorld.getClientPlayer().getFaceDir());		
//		// listen
//		listen();
//		INSPECT(wmWorld.getTime() << "MC: Face after listen: " << wmWorld.getClientPlayer().getFaceDir());				
		
		// if the worldmodel has changed

// 		if ((wmWorld.getTime() != iLastTime
// 		     && (wmWorld.getTime () > lastTurn + 1)) || (wmWorld.getPlayMode() != PLAY_MODE_PLAY_ON))
 		pmLastPlayMode = wmWorld.getPlayMode();
 		if(listen() == 1 || wmWorld.getPlayMode() != pmLastPlayMode) // Got Visual Data
 		{
	 		// log the world
	 		if(fLogging)
	 		    out << wmWorld;
			
//			INSPECT(wmWorld.getTime() << "MC: Face b/f think: " << wmWorld.getClientPlayer().getFaceDir());
			// think
			recFinal = think();
//			INSPECT(wmWorld.getTime() << "MC: Face b/f act: " << wmWorld.getClientPlayer().getFaceDir());
			// act
			act(recFinal);
//			ServerCommand *cmd = act(recFinal);	
			
//			if (cmd != NULL)
//    	    {
//        	    //INSPECT(wmWorld.getTime () << " Command: " << *cmd);
//        		
//    	    	if (cmd->isTurn ()) {
//        	    	INSPECT(wmWorld.getTime () << " Turning!");    		
//        		    lastTurn = wmWorld.getTime ();
//        		}
//        	}
//        		
//    		INSPECT(wmWorld.getTime() << "MC: Face after act: " << wmWorld.getClientPlayer().getFaceDir());
//    		
//    	    iLastTime = wmWorld.getTime();
//		
//		    if (cmd != NULL) delete cmd;
		}
		else
		{
		    bdBody.execute(); // execute whatever is in the queue
//       		if (wmWorld.getTime () == lastTurn + 1)
//       		{
//       		 // send dash
//       		    Recommendation recTemp;
//       		    InterceptAction actTemp(wmWorld);
//       		    actTemp.setDirection(0.0);
//       		    actTemp.setIntensity(50.0);
//       		    recTemp.addAction(actTemp);
//       		    act(recTemp);
//       		
//       		}

		}
		
	}
	
	// close log
	if(fLogging)
	{
    	out << "</playerlog>" << endl;
    	out.close();
    }
}

int MasterControl::listen()
{
	int iVisTime = wmWorld.getVisualTimeStamp();
	
	// get information about the world
	bdBody.senseWorld();
	SensorData* psdData;
	
	// while the body has sensordata
	// TODO: use the importance of the sensordata to triage
	//			 information
	while (bdBody.hasSensorData())
	{
		// grab the next piece of sensordata
		psdData = bdBody.getNextSensorData();		
		// insert the new sensordata into the world
		psdData->worldmodel_insert(wmWorld);

		delete psdData;
		psdData = NULL;
		
    	if(iVisTime != wmWorld.getVisualTimeStamp())
	        return 1;		
	}	
	
	return 0;	
}

Recommendation MasterControl::think()
{
	RecommendationList rclRecommendations;
	Recommendation recTemp;
	
	unsigned int i;
	
	// for each behavior
	for(i = 0; i < pbeBehaviors.size(); i++)
	{
		// if the behavior is useful
		if ((pbeBehaviors[i] != NULL) && pbeBehaviors[i]->isUseful())
		{
			// add it's recommendation to the list
			pbeBehaviors[i]->evaluate(rclRecommendations);
		}
	}
	
	// Log the recommendations
	
	if(fLogging)
	{
// 		if ((wmWorld.getTime() != iLastTime) || (wmWorld.getPlayMode() != PLAY_MODE_PLAY_ON)) {	

    	    // INSPECT(rclRecommendations);
    	    out << rclRecommendations << endl;    	
    	//    	}
	}

	//cerr << "List\n:" << rclRecommendations << "D" << endl;
		
	// return the composition of all of the recommendations
	return rclRecommendations.compose();
}

int MasterControl::act(const Recommendation &_recFinal)
{

  INSPECT ("Playmode: " << wmWorld.getPlayMode());
  INSPECT ("Final: " << _recFinal << "D");
  
  // get the list of actions from the recommendation
  
  ActionList acActions = _recFinal.getActionList();	
  return bdBody.execute(acActions);
}
